#! /usr/bin/env python
# -*- coding:utf-8 -*-

##
## Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
##
## This software is published by the license of CPU-OS License, you can use and
## distibuted this software under this License. See CPU-OS Liense for more detail.
##
## You should have receive a copy of CPU-OS License. If not, please contact us
## by email <support_os@cpu-os.ac.cn>
##
##

import apt, sys, os
import json
import time
import aptdaemon
import aptdaemon.loop
import logging
import commands
import threading
import gettext
from aptdaemon.errors import TransactionFailed, AptDaemonError, NotAuthorizedError
from aptdaemon import client
from subprocess import Popen, PIPE
from threading import Timer
gettext.bindtextdomain('cdosupgrade', "/usr/share/locale/")
gettext.textdomain('cdosupgrade')
_ = gettext.gettext

DKMS_FLAG = False
upgrade_counter = 0
exception_counter = 0
exception_list = []
continue_flag = True
install_counter = 0

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(filename)s %(levelname)s %(message)s',
                    filename='/var/log/upgrade.log',filemode='w')

def test_network():
    result = os.system('ping cdos1.os.cn -c 2')
    if result:
        return False
    else:
        return True
    
def dkms_handle(local_path):
    dkms_file = os.path.join(local_path, ".config/cdosupgrade/dkms")
    cmd = "dkms status > %s" %dkms_file
    os.system(cmd)
    with open(dkms_file, 'r') as f:
        for line in f.readlines():
            package = line.split(',')[0]
            version = line.split(',')[1]
            cmd = "sudo dkms remove -m %s -v %s --all" %(package, version)
            os.system(cmd)
            cmd = "sudo dkms add -m %s -v %s" %(package, version)
            os.system(cmd)
            cmd = "sudo dkms build -m %s -v %s" %(package, version)
            os.system(cmd)
            cmd = "sudo dkms install -m %s -v %s" %(package, version)
            os.system(cmd)
    

#STEP 1:backup system
def backup_system():
    logging.info("开始备份") 
    cmd = ["sudo", "/bin/tar", " cvpzf /backup.tgz --exclude=/proc --exclude=/lost+found --exclude=/backup.tgz --exclude=/home \
    --exclude=/sys --exclude=/mnt --exclude=/media --exclude=/run /"]
    #cmd = ["sudo", "/bin/tar", " cvpzf /backup.tgz /etc"]
    comnd = Popen(' '.join(cmd), shell=True)
    returnCode = comnd.wait()
    logging.info("备份完成") 


#STEP 2: update grub config to recovery system from backup.tgz


#STEP 3: UPDATE APT SOURCES
#---------------------------
#sources_list = os.path.join(os.getenv("HOME"), ".config/cdosupgrade/official-package-repositories.list")
#sources_list = "/usr/lib/cdosupgrade/mintUpdate/rel_upgrades/%s/official-package-repositories.list" % codename
def update_apt_sources(sources_list):
    logging.info("更新源列表") 
    if not os.path.exists(sources_list):
        sys.exit(1)
    if os.path.exists("/etc/apt/sources.list.d/official-source-repositories.list"):
        os.system("rm -fr /etc/apt/sources.list.d/*")
    os.system("cp %s /etc/apt/sources.list.d/official-package-repositories.list" % sources_list)
    os.system("touch 99myown")
    os.system("echo 'APT::Get::AllowUnauthenticated \"true\";'>99myown")
    os.system("cp 99myown /etc/apt/apt.conf.d/")
    os.system("rm 99myown")
    logging.info("更新源列表结束") 


def update_release(local_path):
    ######更新/etc/linuxmint/info 
    logging.info("update release")
    new_info = os.path.join(local_path, ".config/cdosupgrade/info")
    new_lsb_release = os.path.join(local_path, ".config/cdosupgrade/lsb-release")
    new_os_release = os.path.join(local_path, ".config/cdosupgrade/os-release")
    if not (os.path.exists(new_lsb_release) and os.path.exists(new_os_release)):
        logging.info("not new release") 
        sys.exit(1)
    if (os.path.isfile(new_info) and os.path.isdir("/etc/cdos/")):
        logging.info("copy info")
        cmd = "sudo cp %s /etc/cdos/" %new_info
        os.system(cmd)

    logging.info("copy lsb-release")
    cmd = "sudo cp %s /etc/" %new_lsb_release
    os.system(cmd)

    logging.info("copy os-release")
    cmd = "sudo cp %s /etc/" %new_os_release
    os.system(cmd)
    logging.info("update release finished")

def update_filesystem(local_path):
    new_manifest_file = os.path.join(local_path, ".config/cdosupgrade/filesystem.manifest")
    cmd = "sudo cp %s /var/log/installer/" %new_manifest_file
    os.system(cmd)
    
def handle_upload():
    if os.path.exists("/etc/cron.hourly/cdosupload_parhour_post"):
        os.system("sudo rm /etc/cron.hourly/cdosupload_parhour_post")
    os.system("sudo apt-get install cdos-upload")
    cmd = "sed -i 's/alive/upgrade/g' /etc/cron.hourly/cdosupload_parhour_post"
    os.system(cmd)
           
def reboot():
    logging.info("重新启动") 
    os.system("sudo reboot")
    
def upgrade_log(upgrade_sum, install_sum):
    logging.info("本次升级需要升级%d个包" %upgrade_sum)
    logging.info("本次升级需要安装%d个新包" %install_sum)
        
def exception_log():
    global exception_counter
    logging.info("本次升级%d个包未安装成功" %exception_counter)
    for pkg in exception_list:
        logging.info("%s未安装成功" %pkg)
        
def remove_useless_package(local_path):
    remove_file = os.path.join(local_path, ".config/cdosupgrade/filesystem.manifest-remove")
    with open(remove_file, 'r') as f:
        for package in f.readline():
            command = "sudo apt-get remove %s -y" %package
            os.system(command)
            
def get_lastest_version(version_list):
    lastest_version = version_list[0]
    for version in version_list:
        if os.system("dpkg --compare-versions " + version + " gt " + lastest_version) == 0:
            lastest_version = version
    return lastest_version

def get_pkg_version_list(package):
    pkg_version_list = []
    cmd = "apt-cache madison %s|awk -F '|' '{print $2}'" %package
    subp = Popen(cmd, shell=True, stdout=PIPE)
    returncode = subp.wait()
    for version in subp.stdout.read().strip().split('\n'):
        pkg_version_list.append(version.strip())
    return pkg_version_list
    
def auto_config_confirm():
    cmd = "sudo apt-get install python-virtkey"
    os.system(cmd)
    cmd = "sudo apt-get clean"
    os.system(cmd)
    import virtkey               
    while continue_flag:
        v = virtkey.virtkey()
        v.press_keysym(65421)
        v.release_keysym(65421)
        time.sleep(5)
    
             
class AptAction(object):
    def __init__(self, connect = '123', sources_list=None, local_path=None, backup_flag=None):
        self.connect = connect
        self.transaction = None
        self.action = ''
        self.packageName = ''
        self.sources_list = sources_list
        self.local_path = local_path
        self.backup_flag = backup_flag
    
    def init_data(self):
        self.action = ''
        self.packageName = ''
    
    def sendMsg(self, msg , action = 'continue', progress = 0):
        instruction = {'msg': msg, 'action':action, 'progress':progress}
        #logging.info("msg:%s" % msg) 
        if not isinstance(self.connect, str):
            self.connect.send(json.dumps(instruction))
        
    def process(self):
        #更新缓存
        #self.update_cache()     
        upgrade_sum = 0
        install_sum = 0
        upgrade_list = []
        install_list = []
        new_manifest_list = []
        new_manifest = {}       
        # 备份
        if self.backup_flag == "True":
            self.sendMsg(_("Backup, backup cost about ten minutes, please wait patiently..."), 'backup')
            backup_system()
        self.sendMsg(_("Start update..."),'continue')
        # 更新缓存
        os.system("rm -fr /var/cache/apt/archives/*")
        if self.sources_list is not None:
            update_apt_sources(self.sources_list)
        self.update_cache2()
        #self.cache.open()
        #self.cache.upgrade(True)
        #changes = self.cache.get_changes()
        new_manifest_path = os.path.join(self.local_path, ".config/cdosupgrade/filesystem.manifest")
        with open(new_manifest_path) as f:
            for line in f.readlines():
                result = line.strip("\n").split()
                if len(result) >= 2:
                    key = result[0]
                    value = result[1]
                    new_manifest_list.append(key)
                    new_manifest[key] = value
        ############count total package##############
        upgrade_list_path = os.path.join(self.local_path, ".config/cdosupgrade/upgrade_pkgs")
        with open(upgrade_list_path) as f:
            for line in f.readlines():
                upgrade_sum = upgrade_sum + 1
        install_list_path = os.path.join(self.local_path, ".config/cdosupgrade/new_pkgs")
        with open(install_list_path) as f:
            for line in f.readlines():
                install_sum = install_sum + 1
        ##############end count#############
        upgrade_log(upgrade_sum,install_sum)
        total = upgrade_sum + install_sum
        for i in range(1, total+1):
            path = "/var/cache/apt/archives/" + str(i)
            os.mkdir(path)
        ###########################do upgrade and install download###################################
        upgrade_self_list = ["cdos-upgrade", "cdos-upgrade-data", "cdos-update-frontend-gtk", "cdos-update-daemon"]
        remove_file = os.path.join(self.local_path, ".config/cdosupgrade/filesystem.manifest-remove")
        with open(remove_file, 'r') as f:
            for line in f.readlines():
                upgrade_self_list.append(line.strip())
        with open (upgrade_list_path) as f:
            for line in f.readlines():
                package,version = line.strip('\n').split()
                if package in upgrade_self_list:
                    continue
                version_list = get_pkg_version_list(package)
                lastest_version = get_lastest_version(version_list)
                self.upgrade_pkg(package, lastest_version, total)
        with open(install_list_path) as f:
            for line in f.readlines():
                package,version = line.strip('\n').split()
                if package in upgrade_self_list:
                    continue
                version_list = get_pkg_version_list(package)
                lastest_version = get_lastest_version(version_list)
                self.install(package, lastest_version, total)
        ##########################end upgrade and install download###################################
        ##########################do upgrade and install###################################
        #t1 = threading.Thread(target=auto_config_confirm)
        #t1.start()
        self.force_install(total)
        ##########################end do upgrade and install###################################
        #self.sendMsg('升级完毕', 'continue', 100)
        global continue_flag
        continue_flag = False
        #remove_useless_package(self.local_path)
        exception_log()
        update_release(self.local_path)
        update_filesystem(self.local_path)
        #global DKMS_FLAG
        #if DKMS_FLAG:
        #    self.sendMsg("重新安装部分驱动", 'continue', 100)
        #    dkms_handle(self.local_path)
        os.system("sync")
        self.sendMsg(_("Installation completed"), 'quit', 100)
        #time.sleep(1)
        #reboot()     
    
    def update_cache2(self):
        logging.info("update cache start") 
        try:
            self.init_data()
            self.action = 'update'
            self.sendMsg(_("Updating source address cache, please wait patiently..."), 'continue')
            cache = apt.Cache()
            cache.update(sources_list="/etc/apt/sources.list.d/official-package-repositories.list")
            cache.open()
        except (apt.cache.FetchFailedException, apt.cache.FetchCancelledException, apt.cache.CacheClosedException, apt.cache.LockFailedException) as e:
            self.sendMsg(_("Unable to update cache"), 'quit')
            sys.exit(1)
        logging.info("update cache end")        

    def upgrade_pkg(self, package, version, sum):
        try:
            progress = 0
            name = package
            if name.startswith("linux-image"):
                global DKMS_FLAG
                DKMS_FLAG = True
            self.init_data()
            self.action = 'upgrade'
            self.packageName = name
            global upgrade_counter
            upgrade_counter = upgrade_counter + 1
            progress = upgrade_counter * 50/sum
            self.sendMsg(_("Downloading %s...") % name, 'continue', progress)
            self.retry(name, version)
            logging.info("upgrade package: %s, version: %s" % (str(name), str(version)))
            #self.transaction = self.apt_client.upgrade_packages([name])
            #self.transaction.connect("config-file-conflict", self.config_file_conflict)
            #self.transaction.connect('finished', self.finished)
            #self.transaction.connect('error', self.on_error)
            #self.transaction.run()
            #aptdaemon.loop.mainloop.run()
        except(TransactionFailed, AptDaemonError, NotAuthorizedError) as e:
            pass
            #self.sendMsg('无法升级%s' % name, 'continue')
    
    def install(self, package, version, sum):
        try:
            progress = 0
            name = package
            if name.startswith('linux-image'):
                global DKMS_FLAG
                DKMS_FLAG = True
            self.init_data()
            self.action = 'install'
            self.packageName = name
            global upgrade_counter
            upgrade_counter = upgrade_counter + 1
            progress = upgrade_counter * 50/sum
            self.sendMsg(_("Downloading %s...") % name, 'continue', progress)
            self.retry(name, version)
            logging.info("install package: %s, version: %s" % (str(name), str(version)))
            #self.transaction = self.apt_client.install_packages([name])
            #self.transaction.connect("config-file-conflict", self.config_file_conflict)
            #self.transaction.connect("finished", self.finished)
            #self.transaction.connect("error", self.on_error)  
            #self.transaction.run()
            #aptdaemon.loop.mainloop.run()
        except(TransactionFailed, AptDaemonError, NotAuthorizedError) as e:
            pass            
            #self.sendMsg("无法安装%s" % name, 'continue')
         
    def config_file_conflict(self, transaction, cur, new):
        #logging.info("cur : %s, new : %s" % (cur, new))
        transaction.resolve_config_file_conflict(cur, answer = "replace")
    
    def finished(self, transaction, exit_status):
        #logging.info("finished")
        aptdaemon.loop.mainloop.quit()
    
    def on_error(self, transaction, error_code, error_details):
        if self.action == 'install':
            try:
                self.retry(self.packageName)
            except:
                #self.sendMsg("安装%s过程中失败，跳过" % self.packageName, 'continue')
                pass
        elif self.action == 'upgrade':
            try:
                self.retry(self.packageName)
            except:
                pass
                #self.sendMsg("更新%s过程中失败，跳过" % self.packageName, 'continue')
        self.transaction.cancel()
        aptdaemon.loop.mainloop.quit()

    def retry(self, package, version):
        global upgrade_counter
        command = "cd /var/cache/apt/archives/%s && sudo apt-get download %s=%s" %(str(upgrade_counter), package, version)
        os.system(command)

    def force_install(self, total):
        for i in range(1, total+1):
            logging.info("Installing deb in /var/cache/apt/archives/%s/" % str(i))
            deb_name ="/var/cache/apt/archives/%s/*.deb" %str(i)
            kill = lambda process: process.kill()
            cmd = "sudo dpkg --force-all -i %s" % deb_name
            comnd = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
            my_timer = Timer(120, kill, [comnd])
            try:
                my_timer.start()
                returnCode = comnd.wait()
                stdout,stderr = comnd.communicate()
                logging.info("Install OK: Deb in /var/cache/apt/archives/%s/" % str(i))
            except Exception as e:
                logging.info("Error: Deb in /var/cache/apt/archives/%s/" % str(i))
                logging.info("Reason: %s" % str(e))
            finally:
                my_timer.cancel()
            progress = i * 50/total + 50
            self.sendMsg(_("Installing system..."), 'continue', progress)            

if __name__ == '__main__':
    apt_action = AptAction()
    apt_action.process()
    
  
"""    
    def retry(self, package, version):
        kill = lambda process: process.kill()
        cmd = "sudo apt-get install -y %s=%s" %(package, version)
        comnd = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
        my_timer = Timer(180, kill, [comnd])
        try:
            my_timer.start()
            returnCode = comnd.wait()
            stdout,stderr = comnd.communicate()            
        finally:
            my_timer.cancel()             
        global exception_counter
        if int(returnCode) != 0:
            if test_network():
                global exception_counter
                exception_counter = exception_counter + 1
                exception_list.append(package)
                logging.info("%s returncode:%s" %(package, returnCode))
                force_install(package, version)
            else:
                self.sendMsg("网络中断，请重新连接网络继续升级", 'continue')
                time.sleep(10)
                self.retry(package)

"""
