#!/usr/bin/env python
# -*- coding: utf-8 -*-

##
## Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
##
## This software is published by the license of CPU-OS License, you can use and
## distibuted this software under this License. See CPU-OS Liense for more detail.
##
## You should have receive a copy of CPU-OS License. If not, please contact us
## by email <support_os@cpu-os.ac.cn>
##
##

import sys
import dbus
from PyQt4.QtGui import *
from PyQt4.QtCore import *
from screen import Ui_screen, _fromUtf8, _translate
from rotatewidget import RotateWidget
import gettext ,math

import os, apt
import multiprocessing,time
import json

from apt_daemon import AptAction


DEFAULT_STYLE = """
QProgressBar{
background-color: rgb(0, 42, 69);
border: 1px solid rgb(68, 166, 229);
text-align: center;
}

QProgressBar::chunk{
background-color: rgb(68, 166, 229);
margin: 0px;
}
"""

gettext.install("mintupdate", "/usr/share/linuxmint/locale")

def ui_process(conn):
    app = QApplication(sys.argv)    
    widget = Window(conn)
    widget.show()
    QApplication.setOverrideCursor(Qt.BlankCursor)
    app.exec_()
    
def upgrade_process(conn, sources_list, local_path, backup_flag):
    apt_action = AptAction(conn, sources_list, local_path, backup_flag)
    apt_action.process()    


class Window(QWidget):
    def __init__(self, conn, parent=None):
        super(Window, self).__init__(parent)
        self.ui = Ui_screen()
        self.ui.setupUi(self)
        
        self.thread = Thread(conn, self)
        self.setWindowFlags(Qt.X11BypassWindowManagerHint | Qt.WindowStaysOnTopHint)
        #self.setWindowFlags(Qt.WindowMinimizeButtonHint | Qt.WindowCloseButtonHint)
        desktop = QApplication.desktop()
        self.move(0,0)
        self.resize(desktop.size())
        self.progressBar=QProgressBar(self)
        self.progressBar.setGeometry(QRect(300, 850, 720, 20))
        self.progressBar.move((desktop.width()-self.progressBar.width())/2, 850)
        self.progressBar.setStyleSheet(DEFAULT_STYLE)
        #self.progressBar.setStyleSheet(_fromUtf8("FCProgressBar::chunk{background-color:44a6e5; margin:0px;}"))
        self.progressBar.setMinimum(0)
        self.progressBar.setMaximum(100)
        self.progressBar.hide()
        self.ui.label.move((desktop.width()-self.ui.label.width())/2,self.ui.label.y())
        self.ui.label_msg.move((desktop.width()-self.ui.label_msg.width())/2,self.ui.label_msg.y())

        self.rotateWidget = RotateWidget(self)
        self.rotateWidget.move((desktop.width() - self.rotateWidget.width())/2 , 660)
        self.rotateWidget.hide()
               
        QTimer.singleShot(10, self, SLOT("step()"))
        self.connect(self, SIGNAL("backup"), self.backupmode)
        self.connect(self, SIGNAL("updateprogress"), self.setProgress)
        
    def setMsg(self, msg):
        self.ui.label_msg.setText(msg)
        
    def setProgress(self, num):
        self.rotateWidget.hide()
        self.progressBar.show()
        self.progressBar.setValue(int(num))
        
    def backupmode(self):
        self.rotateWidget.show()
    
    @pyqtSlot()
    def step(self):      
        self.thread.start()
        pass

       
class Thread(QThread):
    def __init__(self, conn, parent):
        super(Thread, self).__init__(parent)
        self.window = parent
        self.conn = conn

    def run(self):
        while True:
            msg = self.conn.recv()
            d = json.loads(msg)
            msg = d['msg']
            action = d['action']
            progress = d['progress']
            self.window.setMsg(msg)
            if action == 'backup':
                self.window.emit(SIGNAL("backup"))
            else:    
                self.window.emit(SIGNAL("updateprogress"), progress)
            if action == 'quit':
                self.window.hide()
                QApplication.restoreOverrideCursor()
                QMessageBox.warning(self.window, "警告", msg,
                                   buttons=QMessageBox.Ok,
                                   defaultButton=QMessageBox.NoButton)
                qApp.quit()
                
        
if __name__ == "__main__":
    if os.getuid() != 0:
        print "Run this code as root!"
        sys.exit(1)
    
    if len(sys.argv) != 4:
        print "Missing arguments!"
        sys.exit(1)
    sources_list = sys.argv[1]
    local_path = sys.argv[2]       
    backup_flag = sys.argv[3]
    fix_script = os.path.join(local_path, ".config/cdosupgrade/fix_bug.py")
    if os.path.exists(fix_script):
        cmd = "sudo python %s" %fix_script
        os.system(cmd)
    conn1, conn2 = multiprocessing.Pipe()
    ui_handle = multiprocessing.Process(target=ui_process, args=(conn1,))
    upgrade_handle = multiprocessing.Process(target=upgrade_process, args=(conn2,sources_list,local_path,backup_flag))
    ui_handle.start()
    upgrade_handle.start()
    
